// -*- C++ -*-
/* Copyright (C) 2003 Free Software Foundation, Inc.
 *
 *  Gaius Mulley (gaius@glam.ac.uk) wrote mtsm.h
 *
 *  mtsm.h
 *
 *  provides a mimimal troff state machine which is necessary to
 *  emit meta tags for the post-grotag device driver.
 */

/*
This file is part of groff.

groff is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

groff is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with groff; see the file COPYING.  If not, write to the Free Software
Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#if !defined(MTSM_H)
#define MTSM_H

#include "stringclass.h"

class int_value {
 public:
  int_value ();
 ~int_value ();

  void diff           (FILE *fp, const char *s, int_value compare);
  int  differs        (int_value compare);
  void set            (int v);
  void unset          (void);
  void set_if_unknown (int v);
  int value;
  int is_known;
};

class bool_value : public int_value {
 public:
  bool_value ();
 ~bool_value ();

  void diff           (FILE *fp, const char *s, bool_value compare);
};

class units_value : public int_value {
 public:
  units_value ();
 ~units_value ();

  void diff    (FILE *fp, const char *s, units_value compare);
  int  differs (units_value compare);
  void set     (hunits v);
};

class string_value {
 public:
  string_value ();
 ~string_value ();

  void diff    (FILE *fp, const char *s, string_value compare);
  int  differs (string_value compare);
  void set     (string v);
  void unset   (void);
  string value;
  int is_known;
};

enum bool_value_state   {MTSM_EOL, MTSM_BR, MTSM_SP, LAST_BOOL};
enum int_value_state    {MTSM_FI, MTSM_RJ, MTSM_CE, LAST_INT};
enum units_value_state  {MTSM_IN, MTSM_LL, MTSM_PO, MTSM_TI, LAST_UNITS};
enum string_value_state {MTSM_TA, LAST_STRING};

class statem {
 public:
        statem  ();
        statem  (statem *copy);
       ~statem  ();
 void   flush   (FILE *fp, statem *compare);
 int    changed (statem *s);
 void   merge   (statem *newer, statem *older);

 int issue_no;
 bool_value   bool_values[LAST_BOOL];
 int_value    int_values[LAST_INT];
 units_value  units_values[LAST_UNITS];
 string_value string_values[LAST_STRING];

 void     add_tag (int_value_state t, int v);
 void     add_tag (bool_value_state t);
 void     add_tag (units_value_state t, hunits v);
 void     add_tag (string_value_state t, string v);

 void     sub_tag_ce (void);
 void     add_if_unknown_tag_ce (int v);
 void     add_tag_ta (void);

 string tag;

 void     update (statem *older, statem *newer, int_value_state t);
 void     update (statem *older, statem *newer, bool_value_state t);
 void     update (statem *older, statem *newer, units_value_state t);
 void     update (statem *older, statem *newer, string_value_state t);
};

class stack {
public:
  stack  ();
  stack  (statem *s, stack *n);
  ~stack ();

  stack  *next;
  statem *state;
};

class mtsm {
 public:
          mtsm          ();
         ~mtsm          ();
 void     push_state    (statem *n);
 void     pop_state     (void);
 void     flush         (FILE *fp, statem *s);
 int      changed       (statem *s);
 void     add_tag       (FILE *fp, string s);

 private:
 int      has_changed   (int_value_state t, statem *s);
 int      has_changed   (bool_value_state t, statem *s);
 int      has_changed   (units_value_state t, statem *s);
 int      has_changed   (string_value_state t, statem *s);
 void     inherit       (statem *s, int reset_bool);

 statem *driver;
 stack  *sp;
};

class state_set {
public:
  state_set();
  ~state_set();

  void incl(enum bool_value_state b);
  void incl(enum int_value_state i);
  void incl(enum units_value_state u);
  void incl(enum string_value_state s);

  void excl(enum bool_value_state b);
  void excl(enum int_value_state i);
  void excl(enum units_value_state u);
  void excl(enum string_value_state s);

  int is_in(enum bool_value_state b);
  int is_in(enum int_value_state i);
  int is_in(enum units_value_state u);
  int is_in(enum string_value_state s);

#if 0
  void incl_all(void);
#endif
private:
  int boolset;
  int intset;
  int unitsset;
  int stringset;
};

#endif
